@@ -19,6 +19,7 @@ logger = logging.getLogger(__name__)
 class BatchProcessor:
     def __init__(self, config):
         self.config = config
+        self.do_log = self.config.get("batch-log-logging", True)
         # How many ids to batch simultaneously
         # self.batch_size=self.config.get('batch-log-batch-size', 10)
         # self.item_types=['scrape']
@@ -29,7 +30,7 @@ class BatchProcessor:
         # self.id_source_lock=threading.Lock()
         # How many simultaneous threads will be processing batch items?
         # self.thread_count=1
-        self.task_filter = re.compile("^([0-9a-z\-_]+)$")
+        self.task_filter = re.compile("^([0-9a-z\-_]+(\/[0-9a-z\-_]+)*)$")
         self.last_status_time = None
         self.callables = {}
         self.entry_name = "batch_filter_entrypoint"
@@ -58,6 +59,7 @@ class BatchProcessor:
                 logger.error(f" + '{name}': {error}")
                 message += f"{name}: {error}\n"
             return False, message
+        logger.info(f"Extended logging={self.do_log}")
         return True, ""
 
     def load_module_by_filepath(self, module_name, module_filepath):
@@ -129,8 +131,10 @@ class BatchProcessor:
             for file_object in files_objects:
                 module_filepath = file_object.as_posix()
                 if not module_filepath:
+                    logger.warn(f"Not posix path: {file_object}")
                     continue
                 if file_object.name.startswith("__"):
+                    logger.warn(f"Skipping internal: {file_object}")
                     continue
                 if fk.utils.file_contains_str(module_filepath, self.entry_name):
                     try:
@@ -153,8 +157,8 @@ class BatchProcessor:
                 ret.append((module_filepath, failure))
         return ret
 
-    def get_callable_for_type_raw(self, type):
-        match = self.task_filter.match(type)
+    def get_callable_for_type_raw(self, t):
+        match = self.task_filter.match(t)
         module_name = None
         entry_method = None
         failure = None
@@ -187,20 +191,22 @@ class BatchProcessor:
                     failure = f"No entrypoint found in filter module {module_filepath}"
             else:
                 failure = f"Could not find filter module {module_filepath}"
+        else:
+            failure = f"No match for type {t}"
         return entry_method, failure
 
-    def get_callable_for_type(self, type):
-        if type in self.callables:
-            return self.callables.get(type)
+    def get_callable_for_type(self, t):
+        if t in self.callables:
+            return self.callables.get(t), None
         else:
-            callable = self.get_callable_for_type_raw(type)
-            if callable:
-                self.callables[type] = callable
-            return callable
+            cb, failure = self.get_callable_for_type_raw(t)
+            if cb:
+                self.callables[t] = cb
+            return cb, failure
 
     def _execute_safely(self, entrypoint, item):
         # logger.info("SAFE EXECUTION STARTED!")
-        # We cap runtime at 60 seconds
+        # We cap runtime
         timeout = 30
         failure = None
         result = None
@@ -211,7 +217,7 @@ class BatchProcessor:
             # logger.info("££££ Exit")
         except fk.utils.Watchdog.Watchdog:
             logger.warning("Watchdag triggered exception")
-            failure = f"execution timed out after {timeout} seconds"
+            failure = f"Execution timed out after {timeout} seconds"
         except Exception as e:
             logger.warning("")
             logger.warning("")
@@ -234,13 +240,14 @@ class BatchProcessor:
         """
         item = self.db.book_batch_item(self.db.PENDING, self.db.IN_PROGRESS)
         if item:
-            # logger.info("Processing item:")
-            # logger.info(pformat(item))
+            if self.do_log:
+                logger.info("Processing item:")
+                logger.info(pformat(item))
             id = item.get("id", None)
-            type = item.get("type", None)
+            t = item.get("type", None)
             updated_at = item.get("updated_at", None)
-            if id and type and updated_at:
-                entrypoint, failure = self.get_callable_for_type(type)
+            if id and t and updated_at:
+                entrypoint, failure = self.get_callable_for_type(t)
                 if entrypoint and failure:
                     entrypoint = None
                     failure = f"{failure} AND ENTRYPOINT WAS SET!"
@@ -252,14 +259,58 @@ class BatchProcessor:
                     logger.warning(failure)
                     logger.warning("")
                 id2, updated_at2 = self.db.bump_batch_item(id=id, status=self.db.FAILED if failure else self.db.DONE, error=failure, updated_at=updated_at, result=result) or (None, None)
-                # logger.info(f"id2={id2}, updated_at2={updated_at2}, id={id}, updated_at={updated_at}")
+                if self.do_log:
+                    logger.info(f"id2={id2}, updated_at2={updated_at2}, id={id}, updated_at={updated_at}")
+                return True
             else:
-                logger.warning(f"Missing data for item id={id}, type={type}, updated_at={updated_at}")
-
+                logger.warning(f"Missing data for item id={id}, type={t}, updated_at={updated_at}")
         # else:
         # logger.info(f"No pending items found")
+        return False
+    
+    
 
-    def insert_batch_item(self, type="test", data=None, result=None, priority=50, source=None):
+    def execute_one_throttled_batch_item(self):
+        """
+        Take ownership of one batch item and make sure it is properly executed and updated with status underway
+        NOTE: Throttled version
+        """
+        item = self.db.book_throttled_batch_item(self.db.PENDING, self.db.IN_PROGRESS)
+        if item:
+            if self.do_log:
+                logger.info("Processing item:")
+                logger.info(pformat(item))
+            id = item.get("id", None)
+            t = item.get("type", None)
+            updated_at = item.get("updated_at", None)
+            if id and t and updated_at:
+                if item.get('wait_millis', 0)>0:
+                    logger.warning("Item not booked")
+                    logger.warning(pformat(item))
+                    failure="Not booked"
+                else:
+                    entrypoint, failure = self.get_callable_for_type(t)
+                    if entrypoint and failure:
+                        entrypoint = None
+                        failure = f"{failure} AND ENTRYPOINT WAS SET!"
+                    result = None
+                    if entrypoint:
+                        result, failure = self._execute_safely(entrypoint, item)
+                    if failure:
+                        logger.warning("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤ BATCH FILTER FAILED WITH ¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤")
+                        logger.warning(failure)
+                        logger.warning("")
+                id2, updated_at2 = self.db.bump_batch_item(id=id, status=self.db.FAILED if failure else self.db.DONE, error=failure, updated_at=updated_at, result=result) or (None, None)
+                if self.do_log:
+                    logger.info(f"id2={id2}, updated_at2={updated_at2}, id={id}, updated_at={updated_at}")
+                return True
+            else:
+                logger.warning(f"Missing data for item id={id}, type={t}, updated_at={updated_at}")
+        # else:
+        # logger.info(f"No pending items found")
+        return False
+
+    def insert_batch_item(self, type="test", data=None, result=None, priority=50, source=None, throttle_key=None, throttle_limit=1, throttle_period=1):
         """
         Insert a new batch item into the database, ready for execution
         """
@@ -271,6 +322,9 @@ class BatchProcessor:
                 "result": result,
                 "type": type,
                 "status": self.db.PENDING,
+                "throttle_key": throttle_key,
+                "throttle_limit": throttle_limit,
+                "throttle_period": throttle_period,
                 "source": source
             }
         )
@@ -288,12 +342,23 @@ class BatchProcessor:
     def delete_all_jobs(self, time_sec=30):
         self.db.delete_all()
 
+    def get_job_counts(self):
+        raw = self.db.get_job_counts()
+        out={}
+        for row in raw:
+            t=row.get('type')
+            s=row.get('status')
+            c=row.get('count')
+            out[t]=out.get(t,{})
+            out[t][s]=c
+        return out
+
     def get_status(self):
-        status = {"type": self.db.get_type_counts(), "status": self.db.get_status_counts(), "last": self.db.get_last_jobs(limit=10), "type_status": self.db.get_type_status_counts()}
+        status = {"type": self.db.get_type_counts(), "status": self.db.get_status_counts(), "last": self.db.get_last_jobs(limit=10), "type_status": self.db.get_type_status_counts(), "counts": self.get_job_counts()}
         return status
 
     def process(self):
-        if not self.execute_one_batch_item():
+        if not self.execute_one_throttled_batch_item():
             # Put a status
             # if not self.last_status_time or (datetime.now() - self.last_status_time).total_seconds() > 10.0:
             # 	self.print_status()
