@@ -28,6 +28,9 @@ class Database:
                 result text,
                 type varchar(255),
                 status varchar(255),
+                throttle_key varchar(63),
+                throttle_limit integer,
+                throttle_period integer, 
                 source text,
                 error text,
                 created_at timestamptz not null default now(),
@@ -38,6 +41,9 @@ class Database:
             comment on column batch_log.result is 'The batch item''s result data. Depends entirely on the type. Could for example be the HTML scraped for the input URL for a site_scrape item';
             comment on column batch_log.type is 'The batch item''s type such as order_scrape or site_scrape';
             comment on column batch_log.status is 'The batch item''s status such as pending, in-progress or done';
+            comment on column batch_log.throttle_key is 'When set enables throttling between all items that have the same key';
+            comment on column batch_log.throttle_limit is 'When throttle_key is set, this specifies the number of items that can be processed over throttle_period';
+            comment on column batch_log.throttle_period is 'When throttle_key is set, this specifies the period over which throttle_limit items may be processed';
             comment on column batch_log.source is 'The batch item''s source. Spesifically which component registered it.';
             comment on column batch_log.error is 'The batch item''s error message. Should be None unless status is "failed", in which it case it should be a descriptive error message.';
             comment on column batch_log.created_at is 'When the batch item was first created';
@@ -45,6 +51,67 @@ class Database:
             """
         )
 
+        # Inspired by https://dev.to/astagi/rate-limiting-using-python-and-redis-58gk
+        # Create a function to keep track of throttling of batch items
+        # NOTE: Throttling works by an algorithm called Generic Cell Rate Algoritm (GCRA for short).
+        #       The function is called for a key, and will return an integer number of milliseconds to wait for the key to become unlimited
+        #       The function can be called with do_book=true to actually perform booking if the key is unlimited
+        #       The parameters limit_count and period_millis are used to specify the limit where
+        #       limit_count is the number of actions allowed over the time interval of period_millis
+        self.dbc.query_none(
+            """
+-- function that returns the number of milliseconds to wait before the given throttle_key is available. If the value returned is <=0 that means you just got lucky and no throttle was necessary for that key at this point.
+drop function if exists gcra_throttle (varchar, integer, integer, boolean);
+create or replace function gcra_throttle (throttle_key varchar(63), limit_count integer, period_millis integer, do_book boolean) returns bigint as $$
+declare
+    -- Holds current time in milliseconds since epoch at the entry of the function
+    now_millis bigint;
+    -- Holds our current value of TAT in milliseconds since epoch
+    tat_millis bigint;
+    -- Holds the return value for "how long should we wait before invoking our throttled action?" in milliseconds
+    left_millis bigint;
+begin
+
+    -- Trivial reject: when throttle_key is not set we return 0 to signify no rate limiting needed
+    if (throttle_key = '') is not false then
+        return 0;
+    end if;
+
+    -- Get current time in milliseconds since epoch
+    select extract(epoch from now()) * 1000 into now_millis;
+    
+    -- Make sure our working table is ready
+    create unlogged table if not exists throttle_tat(
+        key varchar(63) primary key,
+        millis bigint
+    );
+
+    -- Hold an exclusive lock on the table to avoid race conditions
+    lock table throttle_tat in exclusive mode;
+
+    -- Get stored value for tat, initializing it to 0 if it was not set 
+    select t.millis into tat_millis from throttle_tat t where t.key = $1;
+    if tat_millis is null then
+        tat_millis := now_millis;
+        insert into throttle_tat (key, millis) values ($1, tat_millis);
+    end if;
+
+    -- Calculate how many milliseconds we should wait before performing our action.
+    -- NOTE: Zero or negative values means we don't need to wait
+    left_millis := (tat_millis - now_millis)  -  (period_millis - (period_millis / limit_count));
+
+    -- We waited long enough, perform our booking if so desired
+    if do_book and left_millis <= 0 then
+        update throttle_tat t set millis = greatest(tat_millis, now_millis) + (period_millis / limit_count) where t.key = $1;
+    end if;
+    -- Tell the caller our status
+    -- return format('now_millis=%L, key=%L, tat=%L, new-tat=%L, separation_millis=%L, do_book=%L, left_millis=%L', now_millis, $1, tat_millis, greatest(tat_millis, now_millis) + (period_millis / limit_count), (period_millis / limit_count), do_book, left_millis);
+    return left_millis;
+end
+$$ language plpgsql;
+"""
+        )
+
     # Look at faster batch imports with
     # from psycopg2.extras import execute_values
     # execute_values(cur,
@@ -63,6 +130,9 @@ class Database:
                     result,
                     type,
                     status,
+                    throttle_key,
+                    throttle_limit,
+                    throttle_period,
                     source
                     )
                 values
@@ -72,6 +142,9 @@ class Database:
                     %(result)s,
                     %(type)s,
                     %(status)s,
+                    %(throttle_key)s,
+                    %(throttle_limit)s,
+                    %(throttle_period)s,
                     %(source)s
                     )
                 on conflict(id)
@@ -81,6 +154,9 @@ class Database:
                     result=%(result)s,
                     type=%(type)s,
                     status=%(status)s,
+                    throttle_key=%(throttle_key)s,
+                    throttle_limit=%(throttle_limit)s,
+                    throttle_period=%(throttle_period)s,
                     source=%(source)s,
                     updated_at=now()
                 returning id
@@ -100,6 +176,9 @@ class Database:
                     result=%(result)s,
                     type=%(type)s,
                     status=%(status)s,
+                    throttle_key=%(throttle_key)s,
+                    throttle_limit=%(throttle_limit)s,
+                    throttle_period=%(throttle_period)s,
                     source=%(source)s,
                     error=%(error)s,
                     updated_at=now()
@@ -120,7 +199,7 @@ class Database:
                     status=%(to_status)s,
                     updated_at=now()
                 where status = %(from_status)s
-                and updated_at < ( now() - (%(time_sec)s * INTERVAL '1 second'))
+                and updated_at < ( now() - (%(time_sec)s * interval '1 second'))
                 ;
                 """,
             {"from_status": (from_status,), "to_status": (to_status,), "time_sec": (time_sec,)},
@@ -159,27 +238,6 @@ class Database:
                 {"id": (id,), "status": (status,), "error": (error,), "result": (result,), "updated_at": (updated_at,)},
             )
 
-    # Delete batch items with given status and update_at longer than given time
-    def delete_batch_items_with_status_older_than(self, status, updated_at):
-        self.dbc.query_none(
-            """
-                delete from batch_log
-                where status=%(status)s,
-                and updated_at >= %(updated_at)s
-                ;
-                """,
-            {"status": (status,), "updated_at": (updated_at,)},
-        )
-
-    # Clear out the batch items table
-    def delete_all(self):
-        return self.dbc.query_none(
-            """
-                delete from batch_log
-                ;
-                """
-        )
-
     # Update a batch item in the batch log, making sure to fail if the id and updated_at don't match, providing a guarantee of atomic operation
     def book_batch_item(self, from_status=PENDING, to_status=IN_PROGRESS):
         return self.dbc.query_one(
@@ -202,8 +260,55 @@ class Database:
             {"from_status": (from_status,), "to_status": (to_status,)},
         )
 
+    # Update a batch item in the batch log, making sure to fail if the id and updated_at don't match, providing a guarantee of atomic operation
+    # NOTE: This is the throttled version
+    def book_throttled_batch_item(self, from_status=PENDING, to_status=IN_PROGRESS):
+        return self.dbc.query_one(
+            """
+                update batch_log
+                    set status=%(to_status)s,
+                    updated_at=now()
+                where status = %(from_status)s
+                and id = (
+                    select id
+                    from batch_log
+                    where true
+                    and status = %(from_status)s
+                    and gcra_throttle(throttle_key, throttle_limit, throttle_period, true) as wait_millis <=0
+                    order by priority desc, updated_at asc, wait_millis desc
+                    limit 1
+                )
+                returning id, priority, data, result, type, status, wait_millis, source, created_at, updated_at
+                ;
+                """,
+            {"from_status": (from_status,), "to_status": (to_status,)},
+        )
+
+
+    # Delete batch items with given status and update_at longer than given time
+    def delete_batch_items_with_status_older_than(self, status, updated_at):
+        self.dbc.query_none(
+            """
+                delete from batch_log
+                where status=%(status)s,
+                and updated_at >= %(updated_at)s
+                ;
+                """,
+            {"status": (status,), "updated_at": (updated_at,)},
+        )
+
+    # Clear out the batch items table
+    def delete_all(self):
+        return self.dbc.query_none(
+            """
+                delete from batch_log
+                ;
+                """
+        )
+
+    
     # Get batch items from batch log fitting the optional filter
-    def get_batch_items(self, id=None, priority=None, type=None, status=None, source=None, limit=1):
+    def get_batch_items(self, id=None, priority=None, type=None, status=None, throttle_key=None, source=None, limit=1):
         return self.dbc.query_many(
             """
                 select
@@ -213,6 +318,9 @@ class Database:
                     result,
                     type,
                     status,
+                    throttle_key,
+                    throttle_limit,
+                    throttle_period,
                     source,
                     created_at,
                     updated_at
@@ -222,12 +330,13 @@ class Database:
                 and (%(priority)s is null or priority = any(%(priority)s))
                 and (%(type)s is null or type = any(%(type)s))
                 and (%(status)s is null or status = any(%(status)s))
+                and (%(throttle_key)s is null or throttle_key = any(%(throttle_key)s))
                 and (%(source)s is null or source = any(%(source)s))
                 order by priority desc, updated_at asc
                 limit %(limit)s
                 ;
                 """,
-            {"id": (id,), "priority": (priority,), "type": (type,), "status": (status,), "source": (source,), "limit": (limit,)},
+            {"id": (id,), "priority": (priority,), "type": (type,), "status": (status,), "throttle_key": (throttle_key,), "source": (source,), "limit": (limit,)},
         )
 
     # Get batch item by id from batch log
@@ -241,6 +350,9 @@ class Database:
                     result,
                     type,
                     status,
+                    throttle_key,
+                    throttle_limit,
+                    throttle_period,
                     source,
                     created_at,
                     updated_at
@@ -292,9 +404,55 @@ class Database:
                 ;
                 """
         )
+    
+    # Get count of ready pending batch items taking into acount throttling
+    def get_throttled_pending_count(self):
+        return self.dbc.query_many(
+            """
+                select
+                    count(*) as count,
+                    min(gcra_throttle(throttle_key, throttle_limit, throttle_period, false)) as wait_millis
+                from batch_log
+                ;
+                """
+        )
+    
+    
+    
+    # Get ready pending batch items taking into acount throttling
+    def get_throttled_pending(self):
+        return self.dbc.query_many(
+            """
+                select
+                    id,
+                    gcra_throttle(throttle_key, throttle_limit, throttle_period, false) as wait_millis
+                from batch_log
+                order by wait_millis asc
+                ;
+                """
+        )
+        
+    def get_job_counts(self):
+        return self.dbc.query_many(
+            """
+                select
+                    type,
+                    status,
+                    count(*)
+                from
+                    batch_log
+                group by
+                    type,
+                    status
+                order by
+                    type,
+                    status
+                ;
+                """
+        )
 
     # Get batch items from batch log sorted by last active
-    def get_last_jobs(self, id=None, priority=None, type=None, status=None, source=None, error=None, limit=1):
+    def get_last_jobs(self, id=None, priority=None, type=None, status=None, throttle_key=None, source=None, error=None, limit=1):
         return self.dbc.query_many(
             """
                 select
@@ -304,6 +462,9 @@ class Database:
                     result,
                     type,
                     status,
+                    throttle_key,
+                    throttle_limit,
+                    throttle_period,
                     error,
                     updated_at - created_at as runtime,
                     extract(EPOCH from (updated_at - created_at)) as runtime_ts,
@@ -315,13 +476,14 @@ class Database:
                 and (%(priority)s is null or priority = any(%(priority)s))
                 and (%(type)s is null or type = any(%(type)s))
                 and (%(status)s is null or status = any(%(status)s))
+                and (%(throttle_key)s is null or throttle_key = any(%(throttle_key)s))
                 and (%(error)s is null or error = any(%(error)s))
                 and (%(source)s is null or source = any(%(source)s))
                 order by updated_at desc, runtime asc, priority desc
                 limit %(limit)s
                 ;
                 """,
-            {"id": (id,), "priority": (priority,), "type": (type,), "status": (status,), "source": (source,), "error": (error,), "limit": (limit,)},
+            {"id": (id,), "priority": (priority,), "type": (type,), "status": (status,), "throttle_key": (throttle_key,), "source": (source,), "error": (error,), "limit": (limit,)},
         )
 
     def get_now(self):
@@ -332,22 +494,3 @@ class Database:
         )
         r = r.replace(tzinfo=pytz.UTC)
         return r
-
-
-def entrypoint(context):
-    db = Database(context.config)
-
-    wi1 = {"data": "bob", "source": "bob", "type": "bob", "status": "bob"}
-    wi2 = {"data": "lol", "source": "lol", "type": "lol", "status": "lol"}
-    id1 = db.insert_batch_item(wi1)
-    id2 = db.insert_batch_item(wi2)
-    # 	type=None, status=None, source=None, limit=None):
-    print("LOL:", db.get_batch_items(type=["lol"]))
-    print("ALL:", db.get_batch_items())
-    print("ONE:", db.get_batch_item_by_id(id1))
-    print("TWO:", db.get_batch_item_by_id(id2))
-
-
-# Main entrypoint of script
-if __name__ == "__main__":
-    entrypoint()
